import os
from abc import ABC, abstractmethod
from argparse import Namespace, ArgumentParser
from functools import partial
from noetl.natstream import NatsStreamReference, NatsPool, NatsConfig, Msg
from aioprometheus import Counter
from aioprometheus.service import Service
from noetl.payload import Payload, logger


class Plugin(NatsPool, ABC):
    events_counter: Counter

    async def subscribe(self, args):
        """
        Default NATS subscription. Can be overridden for custom subscription if needed.
        """
        logger.info(f"Nats stream: {args.nats_subscription_stream}")
        logger.info(f"Nats subject: {args.nats_subscription_subject}")
        cb = partial(self.process_stream, args)
        await self.nats_read(
            subject=args.nats_subscription_subject,
            stream=args.nats_subscription_stream,
            queue=args.nats_subscription_queue,
            cb=cb)

    @abstractmethod
    async def switch(self, payload: Payload):
        """
        Plugin Subclass must implement switch method.
        """
        pass

    async def process_stream(self,
                             args: Namespace,
                             msg: Msg):
        payload = Payload.decode(msg.data)
        payload.set_nats_pool(nats_pool=self.nats_pool)
        payload.info = vars(args)
        payload.nats_reference = NatsStreamReference(
            nats_msg_metadata=msg.metadata,
            nats_msg_subject=msg.subject,
            nats_msg_headers=msg.headers,
            nats_msg_reply=msg.reply
        )
        logger.debug(f"payload: {payload}")
        _ = await self.switch(payload=payload)

    async def run(self, args):
        """
        Default run using NATS. Can be overridden by subclasses.
        """
        service = Service()
        await service.start(addr=args.prom_host, port=args.prom_port)
        logger.info(f"Serving prometheus metrics on: {service.metrics_url}")
        await self.subscribe(args)

    async def shutdown(self):
        if self.nats_pool:
            await self.nats_pool.close_pool()


def parse_args(description, **kwargs):
    parser = ArgumentParser(description=description)
    args = Namespace()
    for arg, value in kwargs.items():
        env_var, default, help_text = value
        env_value = os.getenv(env_var.upper())
        value = env_value if env_value else default
        setattr(args, arg, value)
        parser.add_argument(f"--{arg}", default=value, help=help_text)
    return args



import asyncio
import nats
from nats.js.api import StreamConfig, PubAck, RawStreamMsg
from nats.aio.msg import Msg
from nats.aio.errors import ErrTimeout
from dataclasses import dataclass, asdict
import json
from loguru import logger
from datetime import datetime
from uuid import uuid4
from nats import errors


@dataclass
class NatsConfig:
    nats_url: str
    nats_pool_size: int


@dataclass
class NatsStreamReference:
    nats_msg_subject: str | None = None
    nats_msg_metadata: Msg.Metadata | None = None
    nats_msg_headers: dict | None = None
    nats_msg_reply: Msg.reply = None

    @property
    def __dict__(self):
        return self.to_dict()

    def to_dict(self):
        data = asdict(self)

        def convert_datetime(obj):
            if isinstance(obj, datetime):
                return obj.isoformat()
            elif isinstance(obj, dict):
                return {k: convert_datetime(v) for k, v in obj.items()}
            elif isinstance(obj, list):
                return [convert_datetime(i) for i in obj]
            else:
                return obj

        return convert_datetime(data)

    def to_json(self):
        return json.dumps(self.to_dict())


class NatsConnectionPool:
    _instance = None

    def __new__(cls, config: NatsConfig | None):
        if cls._instance is None:
            cls._instance = super(NatsConnectionPool, cls).__new__(cls)
            cls._instance.config = config
            cls._instance.pool = asyncio.Queue()
        return cls._instance

    @classmethod
    def get_instance(cls, config: NatsConfig | None = None):
        if cls._instance is None:
            cls._instance = cls(config)
        return cls._instance

    async def get(self):
        try:
            if self.pool.empty():
                nc = await nats.connect(self.config.nats_url)
                return nc
            else:
                return await self.pool.get()
        except ErrTimeout as e:
            logger.error(f"NATS connection timeout error: {e}")
            raise
        except Exception as e:
            logger.error(f"NATS connection error: {e}")
            raise

    def get_pool_size(self):
        return self.config.nats_pool_size

    async def put(self, nc):
        try:
            await self.pool.put(nc)
        except Exception as e:
            logger.error(f"NATS connection error to put back to pool: {e}")
            raise

    async def execute(self, func):
        nc = await self.get()
        if nc is None:
            return
        try:
            js = nc.jetstream()
            return await func(js)
        except ErrTimeout as e:
            logger.error(f"NATS JetStream timeout error: {e}")
            raise
        except Exception as e:
            logger.error(f"NATS JetStream error: {e}")
            raise
        finally:
            await self.put(nc)

    def connection(self):
        return self._ConnectionContextManager(self)

    async def publish(self, subject, message, stream) -> PubAck:
        async with self.connection() as js:
            ack: PubAck = await js.publish(subject=f"{subject}", payload=message, stream=stream)
        return ack

    async def truncate_stream(self, stream_name):
        try:
            async with self.connection() as js:
                await js.update_stream(name=stream_name, config=StreamConfig(max_msgs=0))
                print(f"Stream {stream_name} truncated.")
        except Exception as e:
            print(f"Error truncating stream: {e}")

    async def close_pool(self):
        while not self.pool.empty():
            nc = await self.pool.get()
            try:
                await nc.close()
            except Exception as e:
                logger.error(f"NATS connection closing error: {e}")
            else:
                logger.info("NATS connection closed.")
        logger.info("NATS connections in the pool closed.")

    async def bucket_create(self, bucket_name: str):
        async with self.connection() as js:
            try:
                bucket = await js.create_key_value(bucket=bucket_name)
                logger.info(f"Bucket {bucket_name} created.")
                return bucket
            except Exception as e:
                print(f"Bucket create error: {e}")

    async def bucket_delete(self, bucket_name: str):
        async with self.connection() as js:
            try:
                bucket = await js.create_key_value(bucket_name)
                await bucket.delete(bucket_name)
                logger.info(f"Bucket {bucket_name} deleted.")
            except Exception as e:
                print(f"Bucket delete error: {e}")

    async def kv_put(self, bucket_name: str, key: str, value: bytes):
        async with self.connection() as js:
            try:
                kv = await js.create_key_value(bucket=bucket_name)
                revision_number = await kv.put(key, value)
                entry = await kv.get(key)
                logger.debug(f"KeyValue: bucket={bucket_name}, key={entry.key}, revision_number={revision_number}")
                return revision_number
            except Exception as e:
                logger.error(f"Bucket {bucket_name} failed to add kv {key}. Error: {e}")

    async def kv_get(self, bucket_name, key: str):
        async with self.connection() as js:
            try:
                kv = await js.create_key_value(bucket=bucket_name)
                entry = await kv.get(key)
                return entry.value
            except Exception as e:
                logger.error(f"Bucket {bucket_name} failed to get record {key}. Error: {e}")
            return None

    async def kv_get_all(self, bucket_name):
        async with self.connection() as js:
            try:
                kv = await js.create_key_value(bucket=bucket_name)
                keys = await kv.keys()
                return keys
            except Exception as e:
                logger.error(f"Bucket {bucket_name} failed to get keys. Error: {e}")
            return

    async def kv_delete(self, bucket_name, key: str):
        async with self.connection() as js:
            try:
                kv = await js.create_key_value(bucket=bucket_name)
                await kv.delete(key)
                logger.debug(f"Bucket {bucket_name} record {key} deleted")
            except Exception as e:
                logger.error(f"Bucket {bucket_name} failed to delete record {key}. Error: {e}")

    class _ConnectionContextManager:
        def __init__(self, pool):
            self.pool = pool
            self.nc = None

        async def __aenter__(self):
            self.nc = await self.pool.get()
            if self.nc is None:
                raise Exception("Failed to get NATS connection")
            return self.nc.jetstream()

        async def __aexit__(self, exc_type, exc_value, traceback):
            await self.pool.put(self.nc)


class NatsPool:
    nats_pool: NatsConnectionPool

    def __init__(self, nats_pool: NatsConnectionPool | NatsConfig = None):
        if nats_pool:
            self.initialize_nats_pool(nats_pool)

    def initialize_nats_pool(self, nats_pool: NatsConnectionPool | NatsConfig):
        if nats_pool is None:
            self.nats_pool = NatsConnectionPool.get_instance()
        elif isinstance(nats_pool, NatsConfig):
            self.nats_pool = NatsConnectionPool(config=nats_pool)
        elif isinstance(nats_pool, NatsConnectionPool):
            self.nats_pool = nats_pool
        else:
            raise TypeError("nats_pool must be type of NatsConnectionPool or NatsConfig")

    async def get_nats_pool(self):
        if self.nats_pool is None:
            raise ValueError("NatsPool is not initialized")
        else:
            return self.nats_pool

    async def get_msg(self, stream, sequence):
        async with self.nats_pool.connection() as js:
            try:
                msg = await js.get_msg(stream, sequence)
                return msg
            except Exception as e:
                logger.error(f"JetStream get message error for stream {stream} sequence {sequence}: {e}")
                raise

    async def get_consumers_info(self, stream):
        async with self.nats_pool.connection() as js:
            try:
                info = await js.consumers_info(stream)
                return info
            except Exception as e:
                logger.error(f"JetStream get consumer info error for stream {stream}: {e}")
                raise

    async def nats_subject_fetch(self, subject: str, batch=1, timeout=5):
        messages = []

        if self.nats_pool is None:
            await self.get_nats_pool()

        try:
            async with self.nats_pool.connection() as js:
                sub = await js.pull_subscribe(
                    subject=subject,
                    durable=str(uuid4())
                )
                messages += [msg.data for msg in await sub.fetch(batch=batch, timeout=timeout)]
                await sub.unsubscribe()

            return messages

        except ErrTimeout:
            logger.error(f"NATS subject {subject} read timeout error.")
        except Exception as e:
            logger.info(f"NATS subject {subject} read error {e}.")

    async def nats_read(self, subject: str, stream: str, queue: str, cb):
        async with self.nats_pool.connection() as js:
            await js.subscribe(subject=subject, stream=stream, cb=cb, queue=queue)
            while True:
                await asyncio.sleep(1)

    async def nats_write(self, subject: str, stream: str, payload: bytes):
        async with self.nats_pool.connection() as js:
            return await js.publish(subject=subject, stream=stream, payload=payload)


if __name__ == "__main__":
    nats_pool = NatsPool(
        NatsConfig(nats_url="nats://localhost:32222", nats_pool_size=10)
    )


    async def test():
        ack = await nats_pool.nats_write("noetl.command.test", "noetl", b'Hello JS!2')
        print(ack)

        msgs = await nats_pool.nats_subject_fetch("noetl.command.test", batch=100, timeout=1)
        print(msgs)


    async def _main():
        await test()
        nats_pool = NatsConnectionPool.get_instance()
        await nats_pool.close_pool()


    asyncio.run(_main())



import uuid
from datetime import datetime
from enum import Enum
from noetl.keyval import KeyVal
from noetl.const import AppConst
from noetl.natstream import NatsPool, NatsConnectionPool, NatsConfig, NatsStreamReference, ErrTimeout, PubAck, RawStreamMsg, \
    logger

PLAYBOOKS = AppConst.PLAYBOOKS
PLUGINS = AppConst.PLUGINS
PLUGIN_NAME = AppConst.PLUGIN_NAME
IMAGE_URL = AppConst.IMAGE_URL
PLAYBOOK_NAME = AppConst.PLAYBOOK_NAME
PLAYBOOK_BASE64 = AppConst.PLAYBOOK_BASE64
REVISION_NUMBER = AppConst.REVISION_NUMBER
VALUE = AppConst.VALUE
VALUE_NOT_FOUND = AppConst.VALUE_NOT_FOUND
ERROR = AppConst.ERROR
BLUEPRINT = AppConst.BLUEPRINT
BLUEPRINT_SPEC_INPUT = AppConst.BLUEPRINT_SPEC_INPUT
BLUEPRINT_NATS_KV_METADATA = AppConst.BLUEPRINT_NATS_KV_METADATA
PLAYBOOK_INPUT = AppConst.PLAYBOOK_INPUT
METADATA = AppConst.METADATA
METADATA_REFERENCE = AppConst.METADATA_REFERENCE
METADATA_NOT_FOUND = AppConst.METADATA_NOT_FOUND
METADATA_REFERENCE_SUBJECT=AppConst.METADATA_REFERENCE_SUBJECT
METADATA_REFERENCE_CONTEXT=AppConst.METADATA_REFERENCE_CONTEXT
METADATA_REFERENCE_CURRENT_ID=AppConst.METADATA_REFERENCE_CURRENT_ID
METADATA_REFERENCE_ORIGIN_ID=AppConst.METADATA_REFERENCE_ORIGIN_ID
METADATA_REFERENCE_PREVIOUS_ID=AppConst.METADATA_REFERENCE_PREVIOUS_ID
METADATA_REFERENCE_SEQ=AppConst.METADATA_REFERENCE_SEQ
METADATA_REFERENCE_STREAM=AppConst.METADATA_REFERENCE_STREAM
METADATA_REFERENCE_TIMESTAMP=AppConst.METADATA_REFERENCE_TIMESTAMP
NATS_REFERENCE=AppConst.NATS_REFERENCE


EVENT_TYPE=AppConst.EVENT_TYPE
COMMAND_TYPE=AppConst.COMMAND_TYPE
METADATA_EVENT_TYPE=AppConst.METADATA_EVENT_TYPE
METADATA_COMMAND_TYPE=AppConst.METADATA_COMMAND_TYPE

# Events
EVENT_PLUGIN_REGISTERED = AppConst.EVENT_PLUGIN_REGISTERED
EVENT_PLAYBOOK_REGISTERED = AppConst.EVENT_PLAYBOOK_REGISTERED
EVENT_PLAYBOOK_EXECUTION_REQUEST_FAILED = AppConst.EVENT_PLAYBOOK_EXECUTION_REQUEST_FAILED
EVENT_PLAYBOOK_EXECUTION_REGISTERED = AppConst.EVENT_PLAYBOOK_EXECUTION_REGISTERED


class PayloadType(Enum):
    """
    PayloadType Enum class of payload type.
    Attributes:
        EVENT (str): Payload type for events.
        COMMAND (str): Payload type for commands.
    """
    EVENT = AppConst.EVENT
    COMMAND = AppConst.COMMAND


class Payload(KeyVal, NatsPool):
    """The Payload class is used to handle payloads.
    Inherits the KeyVal and NatsPool classes.
    """

    def __init__(self,
                 nats_pool: NatsConnectionPool | NatsConfig | None = None,
                 payload_data: str | dict | list | None = None,
                 event_type: str | None = None,
                 command_type: str | None = None,
                 *args,
                 **kwargs):

        super().__init__(payload_data, *args, **kwargs) if payload_data else super().__init__(self,*args, **kwargs)
        if nats_pool:
            self.initialize_nats_pool(nats_pool)
        self.nats_reference: NatsStreamReference | None = None
        if event_type:
            self.set_event_type(event_type)
        elif command_type:
            self.set_command_type(command_type)

    @property
    def nats_reference(self):
        return self._nats_reference

    @nats_reference.setter
    def nats_reference(self, nats_stream_reference: NatsStreamReference):
        self._nats_reference: NatsStreamReference = nats_stream_reference

    def set_nats_pool(self, nats_pool: NatsConnectionPool | NatsConfig):
        if nats_pool:
            self.initialize_nats_pool(nats_pool)

    def get_reference(self):
        return self.get_value(METADATA_REFERENCE, default=VALUE_NOT_FOUND)

    def get_payload_reference(self):
        return self.get_identifier() | {
            AppConst.SUBJECT: self.get_subject(),
            AppConst.STREAM: self.get_stream(),
            AppConst.SEQ: self.get_seq(),
            AppConst.CONTEXT: self.get_context()
        }

    def get_identifier(self):
        return {
            AppConst.TIMESTAMP: self.get_timestamp(),
            AppConst.CURRENT_ID: self.get_current_id(),
            AppConst.PREVIOUS_ID: self.get_previous_id(),
            AppConst.ORIGIN_ID: self.get_origin_id()
        }

    def get_api_identifier(self) -> str:
        return f"{self.get_origin_id()}.{self.get_previous_id()}.{self.get_current_id()}"

    def get_subject(self):
        return self.get_value(METADATA_REFERENCE_SUBJECT)

    def get_current_id(self) -> str:
        return self.get_value(METADATA_REFERENCE_CURRENT_ID)

    def get_origin_id(self) -> str:
        return self.get_value(METADATA_REFERENCE_ORIGIN_ID)

    def get_previous_id(self) -> str:
        return self.get_value(METADATA_REFERENCE_PREVIOUS_ID)

    def get_timestamp(self) -> str:
        return self.get_value(METADATA_REFERENCE_TIMESTAMP)

    def get_stream(self) -> str:
        return self.get_value(METADATA_REFERENCE_STREAM)

    def get_seq(self) -> str:
        return self.get_value(METADATA_REFERENCE_SEQ)

    def get_context(self) -> str:
        return self.get_value(METADATA_REFERENCE_CONTEXT, AppConst.DEFAULT)

    def get_api_reference(self):
        return {
            AppConst.IDENTIFIER: self.get_api_identifier(),
            AppConst.SUBJECT: self.get_subject(),
            AppConst.STREAM: self.get_stream(),
            AppConst.SEQ: self.get_seq(),
            AppConst.CONTEXT: self.get_context()
        }

    def set_reference(self,
                      reference: dict = None,
                      timestamp=None,
                      subject=None,
                      current_id=None,
                      origin_id=None,
                      previous_id=None,
                      stream=None,
                      seq=None,
                      context=None):
        if not isinstance(reference, dict):
            reference = {}
        self.set_subject(reference.get("subject", subject))
        self.set_timestamp(reference.get("timestamp", timestamp))
        self.set_current_id(reference.get("current_id", current_id))
        self.set_previous_id(reference.get("previous_id", previous_id))
        self.set_origin_id(reference.get("origin_id", origin_id))
        self.set_stream(reference.get("stream", stream))
        self.set_seq(reference.get("seq", seq))
        self.set_context(reference.get("context", context))

    def update_reference(self):
        self.set_previous_id(previous_id=self.get_current_id())
        self.set_current_id()
        self.set_timestamp()

    def set_subject(self, subject: str = None):
        if subject:
            self.set_value(METADATA_REFERENCE_SUBJECT, subject)

    def set_timestamp(self, timestamp: str = None):
        self.set_value(METADATA_REFERENCE_TIMESTAMP, timestamp or str(int(datetime.now().timestamp() * 1000)))

    def set_current_id(self, current_id: str = None):
        self.set_value(METADATA_REFERENCE_CURRENT_ID, current_id or str(uuid.uuid4()))

    def set_previous_id(self, previous_id: str = None):
        self.set_value(METADATA_REFERENCE_PREVIOUS_ID, previous_id or self.get_current_id())

    def set_origin_id(self, origin_id: str = None):
        self.set_value(METADATA_REFERENCE_ORIGIN_ID, origin_id or self.get_current_id())

    def set_stream(self, stream=None):
        if stream:
            self.set_value(METADATA_REFERENCE_STREAM, stream)

    def set_seq(self, seq: str | int = None):
        if seq:
            self.set_value(METADATA_REFERENCE_SEQ, seq)

    def set_context(self, context: str | dict = None):
        if context:
            self.set_value(METADATA_REFERENCE_CONTEXT, context)

    def set_metadata(self, metadata: dict = None, exclude: list[str] = None):
        metadata_orig = self.get_value(METADATA, exclude=exclude)
        if metadata_orig and metadata:
            metadata |= metadata_orig
        self.set_value(METADATA, metadata)

    def add_metadata_value(self, key: str = None, value: any = None):
        if key and value:
            self.set_value(f"{METADATA}.{key}", value)

    def delete_metadata_keys(self, keys: list):
        for key in keys:
            self.delete_value(path=f"{METADATA}.{key}")

    def set_event_type(self, event_type: str = None):
        if event_type:
            self.delete_metadata_keys(keys=list([EVENT_TYPE, COMMAND_TYPE]))
            self.add_metadata_value(EVENT_TYPE, event_type)

    def set_command_type(self, command_type: str = None):
        if command_type:
            self.delete_metadata_keys(keys=list([EVENT_TYPE, COMMAND_TYPE]))
            self.add_metadata_value(COMMAND_TYPE, command_type)

    def set_nats_reference(self):
        self.set_metadata(metadata=self.nats_reference.to_dict(), exclude=NATS_REFERENCE)

    async def snapshot_playbook(self):
        playbook_name = self.get_value(PLAYBOOK_NAME, default=VALUE_NOT_FOUND)
        if playbook_name not in [VALUE_NOT_FOUND, None]:
            kv_payload = await self.playbook_get(playbook_name)
            kv_payload = Payload.decode(kv_payload)
            playbook_template = kv_payload.yaml_value(PLAYBOOK_BASE64)
            if playbook_template == VALUE_NOT_FOUND:
                self.set_value(ERROR, f"Playbook template {playbook_name} was not found")
                self.set_event_type(event_type=EVENT_PLAYBOOK_EXECUTION_REQUEST_FAILED)
            else:
                self.set_value(BLUEPRINT, playbook_template)
                self.set_value(BLUEPRINT_SPEC_INPUT, self.get_value(PLAYBOOK_INPUT))
                self.set_value(BLUEPRINT_NATS_KV_METADATA, kv_payload.get_value(METADATA, default=METADATA_NOT_FOUND))
                self.set_event_type(EVENT_PLAYBOOK_EXECUTION_REGISTERED)

    async def payload_write(self,
                            event_type: str = None,
                            command_type: str = None,
                            message: bytes = None,
                            subject: str = None,
                            subject_prefix: str = None,
                            stream: str = None,
                            plugin: str = None,
                            reference: dict = None):
        if subject_prefix:
            nats_prefix = subject_prefix
        elif self.info:
            prefix_key = "nats_event_prefix" if event_type else "nats_command_prefix"
            nats_prefix = self.info.get(prefix_key, "noetl")
        else:
            nats_prefix = "noetl"

        if event_type:
            self.set_event_type(event_type)
        elif command_type:
            self.set_command_type(command_type)

        reference = reference or self.get_reference()

        if reference not in [VALUE_NOT_FOUND]:
            self.update_reference()
        else:
            self.set_reference()

        subject = subject or f"{nats_prefix}.{plugin}.{self.get_origin_id()}"

        self.set_subject(subject=subject)

        if stream is None:
            stream = self.info.get("nats_subscription_stream")
        self.set_stream(stream=stream)

        if self.nats_reference:
            self.set_nats_reference()

        if message is None:
            message = self.encode()
        ack = await self.nats_write(subject=subject, stream=stream, payload=message)
        logger.debug(ack)
        return ack

    async def command_write(self,
                            command_type: str = None,
                            message: bytes = None,
                            subject: str = None,
                            subject_prefix: str = None,
                            stream: str = None,
                            plugin: str = None,
                            reference: dict = None):

        return await self.payload_write(command_type=command_type,
                                        message=message,
                                        subject=subject,
                                        subject_prefix=subject_prefix,
                                        stream=stream,
                                        plugin=plugin,
                                        reference=reference)

    async def event_write(self,
                          event_type: str = None,
                          message: bytes = None,
                          subject: str = None,
                          subject_prefix: str = None,
                          stream: str = None,
                          plugin: str = None,
                          reference: dict = None):

        return await self.payload_write(event_type=event_type,
                                        message=message,
                                        subject=subject,
                                        subject_prefix=subject_prefix,
                                        stream=stream,
                                        plugin=plugin,
                                        reference=reference)

    async def playbook_bucket_create(self):
        await self.nats_pool.bucket_create(bucket_name=PLAYBOOKS)

    async def playbook_bucket_delete(self):
        await self.nats_pool.bucket_delete(bucket_name=PLAYBOOKS)

    async def playbook_put(self, key: str = None, value: bytes = None):
        revision = 0
        playbook_name = key or self.get_value(PLAYBOOK_NAME)
        playbook = value or self.encode(keys=[PLAYBOOK_BASE64, METADATA])
        if playbook_name and playbook:
            revision = await self.nats_pool.kv_put(bucket_name=PLAYBOOKS, key=playbook_name, value=playbook)
        self.set_event_type(event_type=EVENT_PLAYBOOK_REGISTERED)
        self.set_value(REVISION_NUMBER, revision)

    async def playbook_get(self, key: str):
        return await self.nats_pool.kv_get(bucket_name=PLAYBOOKS, key=key)

    async def playbook_decode(self, key: str):
        kv_payload = await self.playbook_get(key)
        kv_payload_decoded = Payload.decode(kv_payload)
        return kv_payload_decoded.yaml_value(PLAYBOOK_BASE64)


    async def playbook_delete(self, key: str):
        await self.nats_pool.kv_delete(bucket_name=PLAYBOOKS, key=key)

    async def plugin_bucket_create(self):
        await self.nats_pool.bucket_create(bucket_name=PLUGINS)

    async def plugin_bucket_delete(self):
        await self.nats_pool.bucket_delete(bucket_name=PLUGINS)

    async def plugin_put(self, key: str = None, value: bytes = None):
        revision = 0
        plugin_name = key or self.get_value(PLUGIN_NAME)
        plugin = value or self.encode(keys=[PLUGIN_NAME, IMAGE_URL, METADATA])
        if plugin_name and plugin:
            revision = await self.nats_pool.kv_put(bucket_name=PLUGINS, key=plugin_name, value=plugin)
        self.set_event_type(event_type=EVENT_PLUGIN_REGISTERED)
        self.set_value(REVISION_NUMBER, revision)

    async def kv_put_encoded(self, bucket_name: str,  key: str, value: bytes = None):
        revision = await self.nats_pool.kv_put(bucket_name=bucket_name, key=key, value=value or self.encode())
        self.set_value(REVISION_NUMBER, revision)

    async def kv_get_decoded(self, bucket_name: str,  key: str, revision = None):
        try:
            kv = await self.nats_pool.kv_get(bucket_name=bucket_name, key=key)
            if kv:
                return self.decode(kv)
        except Exception as e:
            logger.error(e)
        return None

    async def plugin_get(self, key: str):
        return await self.nats_pool.kv_get(bucket_name=PLUGINS, key=key)

    async def plugin_decode(self, key: str):
        kv_payload = await self.plugin_get(key=key)
        return kv_payload.decode()

    async def plugin_delete(self, key: str):
        await self.nats_pool.kv_delete(bucket_name=PLUGINS, key=key)



import base64
import yaml
import json
from noetl.const import AppConst


class KeyVal(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._info: any = None

    @property
    def info(self):
        return self._info

    @info.setter
    def info(self, value: any):
        self._info = value

    def get_keys(self, path=None) -> list:
        paths = []
        base = self.get_value(path) if path else self
        if isinstance(base, dict):
            for k in base.keys():
                key_path = f"{path}.{k}" if path else k
                paths.append(key_path)
        return paths

    def get_value(self, path: str = None, default: any = None, exclude: list[str] = None):
        exclude = exclude if exclude else []

        def to_dict(obj):
            if isinstance(obj, dict):
                return {k: to_dict(v) for k, v in obj.items() if k not in exclude}
            elif isinstance(obj, list):
                return [to_dict(v) for v in obj]
            else:
                return obj

        if path is None:
            value = to_dict(self)
            return value

        try:
            value = self
            for key in path.split("."):
                if isinstance(value, dict):
                    value = value.get(key)
                elif hasattr(value, key):
                    value = getattr(value, key)
                else:
                    raise TypeError(f"Value for '{key}' is not a dict or does not have attribute '{key}'")
                value = to_dict(value)
                if value is None:
                    return default
            return value
        except Exception as e:
            raise ValueError(f"Error getting value for '{path}': {e}")

    def get_keyval(self, path: str = None, default: any = None, exclude: list[str] = None):
        value = self.get_value(path, default, exclude)
        return KeyVal(value) if isinstance(value, dict) else value

    def set_value(self, path: str, value):
        if path is None:
            raise TypeError("Path cannot be None")
        try:
            keys = path.split(".")
            target = self
            for key in keys[:-1]:
                if key not in target or not isinstance(target[key], dict):
                    target[key] = {}
                target = target[key]

            if isinstance(value, dict) and isinstance(target.get(keys[-1]), dict):
                target[keys[-1]].update(value)
            else:
                target[keys[-1]] = value
        except Exception as e:
            raise ValueError(f"Error setting value for '{path}': {e}")

    def delete_value(self, path: str):
        keys = path.split('.')
        current_key = keys.pop()
        current_dict = self
        for key in keys:
            if key in current_dict:
                current_dict = current_dict[key]
            else:
                return
        if current_key in current_dict:
            del current_dict[current_key]

    def delete_keys(self, keys=None):
        if keys:
            candidates = [k for k in self.get_keys() if k in keys]
            for key in candidates:
                self.delete_value(key)

    def retain_keys(self, keys=None):
        if keys:
            self.delete_keys(keys=[k for k in self.get_keys() if k not in keys])

    def to_json(self):
        try:
            return json.dumps(self.get_value()).encode(AppConst.UTF_8)
        except (TypeError, ValueError) as e:
            raise ValueError(f"Error converting to JSON: {e}")

    def as_json(self, path: str = None, ident: int = None):
        try:
            return json.dumps(self.get_value(path=path), ident=ident)
        except (TypeError, ValueError) as e:
            raise ValueError(f"Error converting to JSON: {e}")

    def base64_path(self, path: str = AppConst.PAYLOAD_BASE64):
        base64_value = self.get_value(path)
        if base64_value is None:
            raise ValueError(f"No base64 string found at {path}")
        if not isinstance(base64_value, str):
            raise TypeError(f"Expected string at '{path}', got {type(base64_value).__name__}")
        return KeyVal.str_base64(base64_value)

    def encode(self, keys=None):
        return base64.b64encode(
            json.dumps(self if keys is None else {key: self[key] for key in keys if key in self.get_value()}).encode(
                AppConst.UTF_8))

    def base64_value(self, path: str = AppConst.VALUE):
        value = self.get_value(path=path, default=AppConst.VALUE_NOT_FOUND)
        if value is None or value == AppConst.VALUE_NOT_FOUND:
            raise ValueError(f"No value found for key {path}")
        elif isinstance(value, str):
            return self.base64_str(value)
        return value

    def yaml_value(self, path: str = AppConst.VALUE):
        value = self.get_value(path=path, default=AppConst.VALUE_NOT_FOUND)
        if isinstance(value, str) and value not in [AppConst.VALUE_NOT_FOUND, None]:
            return self.base64_yaml(value)
        return value

    def yaml_value_dump(self, path: str = AppConst.VALUE):
        return self.yaml_dump(self.yaml_value(path=path))

    @classmethod
    def decode(cls, encoded_payload):
        try:
            payload_data = json.loads(base64.b64decode(encoded_payload).decode(AppConst.UTF_8))
            return cls(**payload_data)
        except Exception as e:
            raise ValueError(f"Error decoding payload: {e}")

    @staticmethod
    def yaml_dump(source: dict):
        return yaml.safe_dump(source, sort_keys=False, allow_unicode=True)

    @staticmethod
    def str_base64(source: str):
        return base64.b64encode(source.encode()).decode(AppConst.UTF_8)

    @staticmethod
    def base64_str(source: str):
        return base64.b64decode(source.encode()).decode(AppConst.UTF_8)

    @staticmethod
    def base64_yaml(source: str):
        try:
            return yaml.safe_load(base64.b64decode(source.encode()).decode(AppConst.UTF_8))
        except yaml.YAMLError as e:
            raise ValueError(f"Error decoding YAML from base64: {e}")

    @classmethod
    def from_json(cls, json_value: str):
        try:
            data = json.loads(json_value)
            return cls(data)
        except json.JSONDecodeError as e:
            raise ValueError(f"Error loading from JSON: {e}")



import unittest
from keyval import KeyVal
import base64
import json

class TestKeyVal(unittest.TestCase):

    def test_get_keys(self):
        kv = KeyVal({"a": 1, "b": 2})
        self.assertListEqual(kv.get_keys(), ["a", "b"])

    def test_get_value(self):
        kv = KeyVal({"a": {"b": {"c": 10}}})
        self.assertEqual(kv.get_value("a.b.c"), 10)
        self.assertIsNone(kv.get_value("a.b.d"))
        self.assertEqual(kv.get_value("a.b.d", default=5), 5)
        with self.assertRaises(TypeError):
            kv.get_value("a.b")

    def test_set_value(self):
        kv = KeyVal()
        kv.set_value("a.b.c", 10)
        self.assertEqual(kv.get_value("a.b.c"), 10)
        with self.assertRaises(TypeError):
            kv.set_value(None, 10)
        with self.assertRaises(ValueError):
            kv.set_value("a.b", [])

    def test_to_json(self):
        kv = KeyVal({"a": 1, "b": 2})
        json_repr = kv.to_json()
        self.assertIsInstance(json_repr, bytes)
        self.assertEqual(json.loads(json_repr.decode('utf-8')), {"a": 1, "b": 2})
        with self.assertRaises(ValueError):
            KeyVal({"a": {1: "invalid"}}).to_json()

    def test_base64_path(self):
        kv = KeyVal({"workflow_base64": base64.b64encode("test".encode()).decode()})
        self.assertEqual(kv.base64_path(), base64.b64encode("test".encode()).decode())
        with self.assertRaises(ValueError):
            kv.base64_path("nonexistent")

    def test_encode(self):
        kv = KeyVal({"a": 1})
        encoded = kv.encode()
        self.assertIsInstance(encoded, bytes)

    def test_yaml_value(self):
        kv = KeyVal({"value": base64.b64encode(json.dumps({"a": 1}).encode()).decode()})
        self.assertEqual(kv.yaml_value(), {"a": 1})

    def test_decode(self):
        kv = KeyVal({"a": 1})
        encoded = kv.encode()
        decoded = KeyVal.decode(encoded)
        self.assertEqual(decoded, {"a": 1})

    def test_base64_str(self):
        self.assertEqual(KeyVal.base64_str("test"), base64.b64encode("test".encode()).decode())

    def test_base64_yaml(self):
        yaml_str = base64.b64encode(json.dumps({"a": 1}).encode()).decode()
        self.assertEqual(KeyVal.base64_yaml(yaml_str), {"a": 1})

    def test_from_json(self):
        json_str = json.dumps({"a": 1})
        kv = KeyVal.from_json(json_str)
        self.assertEqual(kv, {"a": 1})
        with self.assertRaises(ValueError):
            KeyVal.from_json("invalid json")

if __name__ == '__main__':
    unittest.main()



import sys
import requests
import os
import json
from loguru import logger
from noetl.keyval import KeyVal

CWD = os.getcwd()
CSD = os.path.dirname(os.path.abspath(__file__))
API_URL = os.getenv('NOETL_API_URL', "http://localhost:8021/noetl")


class TokenCommand(KeyVal):
    def __init__(self, tokens=None, handler=None, **kwargs):
        super().__init__(**kwargs)
        self.tokens = tokens
        self.handler = handler

    def execute(self):
        if self.handler in ['show_events', 'show_commands']:
            self.graphql_query()
        else:
            mutation = {
                "query": self.create_gql(),
                "variables": self.get_value("variables", {})
            }
            graphql_request(mutation)

    def graphql_query(self):
        query_info = {
            "query": self.create_gql()
        }
        response = requests.post(API_URL, json=query_info)
        if response.status_code == 200:
            print(json.dumps(response.json(), indent=2))
        else:
            print(f"Request failed with status code {response.status_code}")

    def create_gql(self):
        match self.handler:
            case "show_events":
                return """
                query showEvents{
                    showEvents
                }
                    """
            case "show_commands":
                return """
                query showCommands{
                    showCommands
                }
                    """
            case "register_playbook":
                return """
                        mutation RegisterPlaybook($playbookBase64: String!, $metadata: JSON, $tokens: String) {
                            registerPlaybook(playbookBase64: $playbookBase64, metadata: $metadata, tokens: $tokens) {
                                reference 
                                name
                                eventType
                                status
                                message
                            }
                        }
                        """
            case "register_plugin":
                return """
                        mutation RegisterPlugin($registrationInput: PluginRegistrationInput!) {
                          registerPlugin(registrationInput: $registrationInput) {
                            reference 
                            name
                            eventType
                            status
                            message
                          }
                        }
                """
            case "list_playbooks":
                return """
                query ListPlaybooks {
                    listPlaybooks
                }
                """
            case "list_plugins":
                return """
                query ListPlugins {
                    listPlugins
                }
                """
            case "delete_events":
                return """
                mutation DeleteEvents {
                    deleteEvents {
                        message
                    }
                }
                """
            case "delete_commands":
                return """
                mutation DeleteCommands {
                    deleteCommands {
                        message
                    }
                }
                """
            case "describe_playbook":
                return """
                query DescribePlaybook($playbookInput: DescribePlaybookInput!) {
                  describePlaybook(playbookInput: $playbookInput)
                }
                """
            case "describe_plugin":
                return """
                query DescribePlugin($pluginInput: DescribePluginInput!) {
                  describePlugin(pluginInput: $pluginInput)
                }
                """
            case "run_playbook":
                return """
                query RunPlaybook($runPlaybookInput: RunPlaybookInput!) {
                  runPlaybook(runPlaybookInput: $runPlaybookInput) {
                    reference
                    name
                    eventType
                    status
                    message
                  }
                }
                """
            case _:
                raise NotImplementedError(f"Mutation for tokens '{self.tokens}' does not exists.")

    @classmethod
    def create(cls, tokens):
        logger.info(tokens)
        tokens_list = tokens.split(" ")
        if tokens_list[0].lower() == "exit":
            logger.info("\nGoodbye...")
            sys.exit()
        elif len(tokens_list) > 1:
            handler = f"{tokens_list[0]}_{tokens_list[1]}"
            args = tokens_list[2:]
            match handler:
                case "register_playbook":
                    with open(args[0], 'r') as file:
                        return cls(
                            tokens=tokens,
                            handler=handler,
                            variables={
                                "playbookBase64": cls.str_base64(file.read()),
                                "metadata": {"source": "noetl-cli", "handler": handler},
                                "tokens": tokens
                            }
                        )
                case "register_plugin":
                    if len(args) == 2:
                        return cls(
                            tokens=tokens,
                            handler=handler,
                            variables={"registrationInput":
                                {
                                    "pluginName": args[0], "imageUrl": args[1],
                                    "metadata": {"source": "noetl-cli", "handler": handler},
                                    "tokens": tokens
                                }
                            }
                        )
                case "list_playbooks":
                    return cls(
                        tokens=tokens,
                        handler=handler
                    )
                case "list_plugins":
                    return cls(
                        tokens=tokens,
                        handler=handler
                    )
                case "describe_playbook":
                    if len(args) == 1:
                        return cls(
                            variables={"playbookInput": {"playbookName": args[0]}},
                            tokens="describe playbook",
                            handler="describe_playbook"
                        )
                case "run_playbook":
                    if len(args) >= 2:
                        try:
                            playbook_input = json.loads(args[1])
                            playbook_input_json = json.dumps(playbook_input)
                        except json.JSONDecodeError:
                            raise ValueError("Invalid JSON for playbookInput")

                        return cls(
                            variables={"runPlaybookInput":
                                           {"playbookName": args[0],
                                            "input": playbook_input_json
                                            }
                                       },
                            tokens="run playbook",
                            handler="run_playbook"
                        )
                    else:
                        raise ValueError("Need playbookName, playbookInput arguments to run playbook")
                case "describe_plugin":
                    if len(args) == 1:
                        return cls(
                            variables={"pluginInput": {"pluginName": args[0]}},
                            tokens="describe plugin",
                            handler="describe_plugin"
                        )
                case "delete_events":
                    return cls(
                        payload={},
                        metadata={},
                        tokens="delete_events",
                        handler="delete_events"
                    )
                case "delete_commands":
                    return cls(
                        payload={},
                        metadata={},
                        tokens="delete commands",
                        handler="delete_commands"
                    )
                case "show_events":
                    return cls(
                        tokens=tokens,
                        handler=handler,
                    )
                case "show_commands":
                    return cls(
                        tokens=tokens,
                        handler=handler,
                    )
        raise ValueError(f"Unknown command: {tokens}")


def graphql_request(mutation):
    try:
        response = requests.post(API_URL, json=mutation)
        response.raise_for_status()
        if response.ok:
            if "describePlaybook" in response.json().get('data', {}):
                response_keyval = KeyVal(response.json())
                print(response_keyval.base64_value("data.describePlaybook.playbook.playbookBase64"))
            else:
                print(json.dumps(response.json(), indent=2))
        else:
            print(response)
    except requests.exceptions.RequestException as e:
        logger.error(f"API request error: {e}")
        return None


def is_api_running():
    try:
        response = requests.get(f"{API_URL}/health")
        return response.status_code == 200
    except requests.exceptions.RequestException as e:
        logger.error(f"NoETL API health check failed: {e}")
        return False


def main():
    """
    NoETL command line interface.
    Usage:
      - To register a playbook: python noetl/cli.py register playbook "<path_to_playbook_yaml>"
      - To register a plugin: python noetl/cli.py register plugin "<plugin_name>" "<image_url>"
      - To exit the CLI: Type 'exit' when prompted for a command.
    The NoETL CLI supports registering playbooks and plugins by sending GraphQL mutations to the NoETL API endpoint.
    Set the NoETL API endpoint with an environment variable, e.g., NOETL_API_URL="http://localhost:8021/noetl".
    """
    try:
        if len(sys.argv) > 1:
            TokenCommand.create(" ".join(sys.argv[1:])).execute()
        else:
            logger.info("NoETL command line terminal.\nEnter a <command> or 'exit' to quit")
            while True:
                line = input("> ").lstrip()
                if line.lower() == "exit":
                    logger.info("\nGoodbye...")
                    break
                try:
                    TokenCommand.create(line).execute()
                except ValueError as e:
                    logger.error(f"Invalid command: {e}")
                except FileNotFoundError:
                    logger.error(f"File not found")
                except NotImplementedError as e:
                    logger.error(f"Not implemented: {e}")

    except KeyboardInterrupt:
        logger.info("\nGoodbye...")
    except Exception as e:
        logger.error(f"Unexpected error: {str(e)}")


if __name__ == "__main__":
    main()



import uuid
from noetl.natstream import NatsConnectionPool, NatsConfig
from noetl.payload import Payload, AppConst,  RawStreamMsg


class Playbook(Payload):
    def __init__(self,
                 playbook_template: dict | None = None,
                 playbook_id: str | None = None,
                 playbook_input: dict | None = None,
                 playbook_metadata: dict | None = None,
                 nats_pool: NatsConnectionPool | NatsConfig = None, **kwargs):
        if playbook_template:
            kwargs = kwargs | playbook_template
        super().__init__(nats_pool=nats_pool, **kwargs)
        if playbook_id:
            self.set_value("spec.reference.origin_id", playbook_id)
        if playbook_input:
            self.set_value("spec.input", playbook_input)
        if playbook_metadata:
            self.set_value("metadata.nats.kv.metadata", playbook_metadata)

    # def execution_tree(self):
    #     tasks = self.get_keys(path="spec.tasks")
    #     for task_path in tasks:
    #         task_id = str(uuid.uuid4())
    #         self.set_value(f"{task_path}.id", task_id)
    #
    #         steps = self.get_keys(path=f"{task_path}.steps")
    #         for step_path in steps:
    #             step_id = str(uuid.uuid4())
    #             self.set_value(f"{step_path}.id", step_id)

    async def register(self,subject: str, stream: str):
        #subject = f"playbook.{self.get_value('spec.id')}"
        ack = await self.event_write(
            message=self.encode(),
            stream=stream,
            subject=subject
        )
        return ack

    async def generate_command(self):
        pass


    @classmethod
    def unmarshal(cls, binary_data: bytes, nats_pool: NatsConnectionPool | NatsConfig = None):
        return cls(nats_pool=nats_pool, **cls.decode(binary_data))



class AppConst:
    BASE64 = "base64"
    BLUEPRINT = "blueprint"
    BLUEPRINT_SPEC_INPUT = "blueprint.spec.input"
    BLUEPRINT_NATS_KV_METADATA = "blueprintNatsKvMetadata"
    COMMAND = "command"
    COMMAND_TYPE = "commandType"
    CONTEXT = "context"
    CURRENT_ID = "currentId"
    DATA = "data"
    DEFAULT = "default"
    DICT = "dict"
    DISPATCHER = "dispatcher"
    EVENT = "event"
    EVENT_TYPE = "eventType"
    ERROR = "error"
    HTTP_HANDLER = "http_handler"
    IDENTIFIER = "identifier"
    IMAGE_URL = "imageUrl"
    INPUT = "input"
    METADATA = "metadata"
    METADATA_NOT_FOUND = "metadataNotFound"
    NAME = "name"
    NATS_REFERENCE = "natsReference"
    NO_DATA_PROVIDED = "noDataProvided"
    ORIGIN_ID = "originId"
    OUTPUT = "output"
    PAYLOAD_BASE64 = "payloadBase64"
    PAYLOAD_REFERENCE = "payloadReference"
    PLAYBOOK = "playbook"
    PLAYBOOK_BASE64 = "playbookBase64"
    PLAYBOOK_INPUT = "playbookInput"
    PLAYBOOK_NAME = "playbookName"
    PLAYBOOK_METADATA = "playbookMetadata"
    PLAYBOOKS = "playbooks"
    PLAYBOOK_REFERENCE = "playbookReference"
    PLUGIN = "plugin"
    PLUGIN_NAME = "pluginName"
    PLUGINS = "plugins"
    PREVIOUS_ID = "previousId"
    REGISTRAR = "registrar"
    REFERENCE = "reference"
    REVISION_NUMBER = "revisionNumber"
    SEQ = "seq"
    STREAM = "stream"
    SUBJECT = "subject"
    TIMESTAMP = "timestamp"
    TOKENS = "tokens"
    UTF_8 = "utf-8"
    VALUE = "value"
    VALUE_NOT_FOUND = "valueNotFound"
    VALUE_TYPE = "valueType"

    # Metadata

    METADATA_REFERENCE = "metadata.reference"
    METADATA_REVISION_NUMBER = "metadata.revisionNumber"
    METADATA_TOKENS = "metadata.tokens"
    METADATA_PLAYBOOK_NAME = "metadata.playbookName"
    METADATA_EVENT_TYPE = "metadata.eventType"
    METADATA_COMMAND_TYPE = "metadata.commandType"
    METADATA_NAME = "metadata.name"
    METADATA_REFERENCE_CONTEXT = "metadata.reference.context"
    METADATA_REFERENCE_CURRENT_ID = "metadata.reference.currentId"
    METADATA_REFERENCE_ORIGIN_ID = "metadata.reference.originId"
    METADATA_REFERENCE_PREVIOUS_ID = "metadata.reference.previousId"
    METADATA_REFERENCE_SEQ = "metadata.reference.seq"
    METADATA_REFERENCE_STREAM = "metadata.reference.stream"
    METADATA_REFERENCE_SUBJECT = "metadata.reference.subject"
    METADATA_REFERENCE_TIMESTAMP = "metadata.reference.timestamp"

    # Spec
    SPEC_ID = "spec.id"
    SPEC_REFERENCE = "spec.reference"
    SPEC_REFERENCE_SEQ = "spec.reference.seq"
    SPEC_REFERENCE_STREAM = "spec.reference.stream"
    SPEC_INPUT = "spec.input"

    # event types

    EVENT_PLAYBOOK_EXECUTION_REQUEST_FAILED = "PlaybookExecutionRequestFailed"
    EVENT_PLAYBOOK_REGISTRATION_REQUESTED = "PlaybookRegistrationRequested"
    EVENT_PLAYBOOK_EXECUTION_REQUESTED = "PlaybookExecutionRequested"
    EVENT_PLUGIN_REGISTRATION_REQUESTED = "PluginRegistrationRequested"
    EVENT_PLAYBOOK_REGISTERED = "PlaybookRegistered"
    EVENT_PLUGIN_REGISTERED = "PluginRegistered"
    EVENT_PLAYBOOK_EXECUTION_REGISTERED = "PlaybookExecutionRegistered"

    # command types

    COMMAND_REGISTER_PLAYBOOK = "RegisterPlaybook"
    COMMAND_REGISTER_PLUGIN = "RegisterPlugin"
    COMMAND_REGISTER_PLAYBOOK_EXECUTION = "RegisterPlaybookExecution"



from setuptools import setup, find_packages
from setuptools.command.install import install as InstallCommand
import os


def read_requirements():
    with open('requirements.txt', 'r') as f:
        return f.read().splitlines()


def read_readme():
    this_directory = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
        return f.read()


setup(
    name="noetl",
    version="0.1.13",
    author="NoETL Team",
    description="NoETL: A Python package for managing workflows",
    long_description=read_readme(),
    long_description_content_type='text/markdown',
    packages=find_packages(),
    install_requires=read_requirements(),
    entry_points={
        "console_scripts": [
            "noetl = noetl.cli:main",
        ],
    },
)
